# Those constants are used from the library only
BASE_PAL_URL = "https://pal-{}.adyen.com/pal/servlet"
BASE_HPP_URL = "https://{}.adyen.com/hpp"
ENDPOINT_CHECKOUT_TEST = "https://checkout-test.adyen.com"
ENDPOINT_CHECKOUT_LIVE_SUFFIX = "https://{}-checkout-live" \
                                ".adyenpayments.com/checkout"
API_CHECKOUT_VERSION = "v41"
API_CHECKOUT_UTILITY_VERSION = "v1"
API_RECURRING_VERSION = "v25"
API_PAYMENT_VERSION = "v40"
API_PAYOUT_VERSION = "v30"
LIB_VERSION = "2.0.0"
LIB_NAME = "adyen-python-api-library"
