from setuptools import setup

setup(
    name='Adyen',
    packages=['Adyen'],
    version='2.0.0',
    maintainer='Adyen',
    maintainer_email='support@adyen.com',
    description='Adyen Python Api',
    long_description="A Python client library for accessing Adyen APIs",
    author='Adyen',
    author_email='support@adyen.com',
    url='https://github.com/Adyen/adyen-python-api-library',
    keywords=['payments', 'adyen', 'fintech'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6'
    ]
)
